using System;
using System.Collections;
using System.Data;
using System.Globalization;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using Macaw.ReportingServices.SharepointListExtension.WssSite.Lists;
using Macaw.ReportingServices.SharepointListExtension.WssSite.Views;
using Microsoft.ReportingServices.DataProcessing;
using CommandBehavior = Microsoft.ReportingServices.DataProcessing.CommandBehavior;
using CommandType = Microsoft.ReportingServices.DataProcessing.CommandType;
using data = System.Data;
using IDataParameter = Microsoft.ReportingServices.DataProcessing.IDataParameter;
using IDataParameterCollection = Microsoft.ReportingServices.DataProcessing.IDataParameterCollection;
using IDataReader = Microsoft.ReportingServices.DataProcessing.IDataReader;
using IDbCommand = Microsoft.ReportingServices.DataProcessing.IDbCommand;
using IDbTransaction = Microsoft.ReportingServices.DataProcessing.IDbTransaction;

namespace Macaw.ReportingServices.SharepointListExtension
{
	/// <summary>
	/// The Command class does the actual work. It can retrieve meta info from the SharePoint webservices and can
	/// download the list entries. These are used to populate a DataTable that is passed to the DataReader object.
	/// </summary>
	public class Command : Microsoft.ReportingServices.DataProcessing.IDbCommand, IDbCommandAnalysis
	{
		Connection _connection;
		DataParameterCollection _parameters = new DataParameterCollection();
     
		public Command()
		{
			// Implement the default constructor here
		}

		// Implement other constructors here.
		public Command(Connection connection)
		{
			_connection = connection;
		}

		public Command(string cmdText)
		{
			CommandText = cmdText;
		}

		public Command(string cmdText, Connection connection)
		{
			CommandText    = cmdText;
			_connection  = connection;
		}

		/// <summary>
		/// Implements IDbCommandAnalysis.GetParameters
		/// GetParameters is called by the Report Designer
		/// </summary>
		/// <returns></returns>
		public IDataParameterCollection GetParameters ()
		{
			DataParameterCollection FixedParams = new DataParameterCollection();
			IDataParameter SiteUrl = CreateParameter();
			SiteUrl.ParameterName = "SiteUrl";
			SiteUrl.Value = "";
            FixedParams.Add(SiteUrl);
			return FixedParams;
		}

		string _cmdText;
		Hashtable _cmdTextParts = new Hashtable();

		public string CommandText
		{
			get { return _cmdText;  }
			set  
			{
				_cmdText = value;  
				foreach(string line in value.Split(';'))
				{
					if(line.IndexOf("=") > 0)
					{
						_cmdTextParts[line.Substring(0, line.IndexOf("=")).Trim().ToLower()] = line.Substring(line.IndexOf("=") + 1).Trim();
					}
					else if(line.Trim() != "")
					{
						_cmdTextParts["list"] = line.Trim();
					}
				}
			}
		}
		public string ListName
		{
			get{return (string)_cmdTextParts["list"];}
		}

		public string ViewName
		{
			get{return (string)_cmdTextParts["view"];}
		}

		public bool StripHtml
		{
			get
			{
				if(_cmdTextParts["striphtml"] == null)return false;
				return (((string)_cmdTextParts["striphtml"]).ToLower() == "true");
			}
		}

		public TimeGrid Grid
		{
			get
			{
				string gridString = (string)_cmdTextParts["grid"];
				if (gridString == null)return TimeGrid.None;
				try
				{
					return (TimeGrid)Enum.Parse(typeof(TimeGrid), gridString, true);
				}
				catch
				{
					throw new ApplicationException(gridString + " is not a recognized frequency. Use Daily, Weekly, Monthly or Auto");
				}
			}
		}


		private TimeGrid _calcGrid = TimeGrid.None;
		protected TimeGrid RuntimeGrid
		{
			get
			{
				TimeGrid FixedValue = this.Grid;
				if(FixedValue == TimeGrid.Auto)
				{
                    return	_calcGrid;				
				}else
				{
					return FixedValue;
				}
			}
		}

		public DateTime StartDate
		{
			get
			{
				string dateString = (string)_cmdTextParts["startdate"];
				if (dateString == null)return DateTime.MinValue;

				return DateTime.ParseExact(dateString, "yyyy-MM-dd", CultureInfo.InvariantCulture);
			}	
		}
		public DateTime EndDate
		{
			get
			{
				string dateString = (string)_cmdTextParts["startdate"];
				if (dateString == null)return DateTime.MinValue;

				return DateTime.ParseExact(dateString, "yyyy-MM-dd", CultureInfo.InvariantCulture);
			}	
		}


		public int CommandTimeout
		{
			get  { return 300000; }
			set  { }
		}


		private CommandType _type = CommandType.Text;
		/// <summary>
		/// We only use the commandType CommandType.Text
		/// </summary>
		public CommandType CommandType
		{
			get { return _type; }
			set 
			{ 
				if (value != CommandType.Text) throw new NotSupportedException(); 
				_type = value;
			}
		}


		public DataParameterCollection Parameters
		{
			get  { return _parameters; }
		}

		IDataParameterCollection IDbCommand.Parameters
		{
			get  { return _parameters; }
		}

		/// <summary>
		/// We do not support transactions
		/// </summary>
		public IDbTransaction Transaction
		{

			get { return null; }
			set {  }
		}

		/// <summary>
		/// This extension does not support canceling a command
		/// once it has been initiated.
		/// </summary>
		public void Cancel()
		{
			throw new NotSupportedException();
		}

		/// <summary>
		/// Factory method for creating the correct IDbParameter type
		/// </summary>
		/// <returns></returns>
		public IDataParameter CreateParameter()
		{
			return new DataParameter();
		}

		/// <summary>
		/// This is the method that is executed when the
		/// </summary>
		/// <returns></returns>
		public IDataReader ExecuteReader()
		{
			return ExecuteReader(CommandBehavior.SingleResult);
		}

		public IDataReader ExecuteReader(CommandBehavior behavior)
		{
			try
			{
				string SiteUrl = this._connection.WssSiteUrl;
				string s = this.Parameters.SiteUrl;
				if(s != "")
				{
					SiteUrl = s;
				}
				DataTable table = GetDataTableFromWSS(SiteUrl, ListName, ViewName, (behavior == CommandBehavior.SchemaOnly));
				return new DataReader(this._connection, table, _parameters);
			}
			catch (Exception ex)
			{
				System.Diagnostics.Debug.WriteLine(ex.Message);
                throw;
			}

		}

		public void Dispose() 
		{
		}



		/// <summary>
		/// This method calls the web services and fills a data table with the results
		/// </summary>
		/// <param name="site"></param>
		/// <param name="listName"></param>
		/// <returns></returns>
		private DataTable GetDataTableFromWSS(string site, string listName, string viewName, bool onlySchema)
		{

			Lists lists = new Lists();
			lists.Url = site + "_vti_bin/Lists.asmx";
			lists.Credentials = GetCredential(_connection);
			lists.PreAuthenticate = true;
			lists.Timeout = this.CommandTimeout;

			Views views = new Views();
			views.Url = site + "_vti_bin/Views.asmx";
			views.Credentials = GetCredential(_connection);
			views.PreAuthenticate = true;
			views.Timeout = this.CommandTimeout;

			string TechListName = FindTechListName(listName, lists, site);
			string TechViewName = FindTechViewName(viewName, TechListName, views, site);
			DataTable result = new DataTable("list");

			XmlNode ListInfoNode = lists.GetList(TechListName);
			StringBuilder fieldRefs = new StringBuilder();

			Hashtable DisplayNames = new Hashtable();
			Hashtable HtmlFields = new Hashtable();
			foreach(XmlElement Field in ListInfoNode.SelectNodes("wss:Fields/wss:Field", NameSpaceMgr))
			{
				string FieldName = Field.GetAttribute("Name");
				string FieldDisplayName = Field.GetAttribute("DisplayName");
				
				// Computed fields sometimes have Displaynames that are also used by other fields
				if(Field.GetAttribute("Type") == "Computed")
				{
					FieldDisplayName = FieldDisplayName + " (" + FieldName + ")";
				}

                int suffix = 1;
                while (result.Columns.Contains(FieldDisplayName))
                {
                    FieldDisplayName = Field.GetAttribute("DisplayName") + "_" + suffix.ToString();
                    suffix++;
                }

				result.Columns.Add(FieldDisplayName, TypeFromField(Field));
				fieldRefs.AppendFormat("<FieldRef Name=\"{0}\"/>", FieldName);
				if(Field.GetAttribute("RichText") == "TRUE")
				{
					HtmlFields.Add(FieldDisplayName, true);
				}
				DisplayNames.Add(FieldDisplayName, FieldName);
			}

			if(onlySchema)
			{
				if(Grid != TimeGrid.None)
				{
					if(!result.Columns.Contains("Current"))
					{
						throw new ApplicationException("You cannot use Time Grids on lists that don't have a history. Create lists based upon Issue to have historical information.");	
					}
					AddGridDate(result);
				}
				return result;
			}

			XmlElement fields = ListInfoNode.OwnerDocument.CreateElement("ViewFields");
			fields.InnerXml = fieldRefs.ToString();

			XmlElement query = CreateWhereClauseFromView(ListInfoNode, TechListName, TechViewName, views);

            XmlNode ItemsNode = lists.GetListItems(TechListName, TechViewName, query, fields, "100000", null);
            //XmlNode ItemsNode = lists.GetListItems(TechListName, null, null, null, null, null);

			// Lookup fields always start with the numeric ID, then ;# and then the string representation. 
			// We are normally only interested in the name, so we strip the ID.
			// Numeric calculated fields have string type and look like float;#9.000
			Regex CheckLookup = new Regex("^\\w+;#");


			System.Globalization.CultureInfo oldCulture = System.Threading.Thread.CurrentThread.CurrentCulture;
			System.Threading.Thread.CurrentThread.CurrentCulture = System.Globalization.CultureInfo.InvariantCulture;

			foreach(XmlElement Item in ItemsNode.SelectNodes("rs:data/z:row", NameSpaceMgr))
			{
				DataRow newRow = result.NewRow();
				foreach(DataColumn col in result.Columns)
				{
					if(Item.HasAttribute("ows_" + (string)DisplayNames[col.ColumnName]))
					{
						object val = Item.GetAttribute("ows_" + (string)DisplayNames[col.ColumnName]);
						if(val is string && CheckLookup.IsMatch((string)val))
						{
							string valString = val as String;
							val = valString.Substring(valString.IndexOf("#") + 1);
						}
						if(StripHtml && val is string && HtmlFields[col.ColumnName] != null)
						{
							val = ConvertToText((string)val);
						}
						newRow[col] = val;
					}
				}
				result.Rows.Add(newRow);
			}

			System.Threading.Thread.CurrentThread.CurrentCulture = oldCulture;

			if(Grid != TimeGrid.None )
			{
				result = ConvertToTimeGrid(result);
			}

			return result;

		}

		/// <summary>
		/// For simple lists (without view specified), we want all records, even if the default view has constraints,
		/// for other views, we want to remove the constraint on Current if Grid != None
		/// </summary>
		/// <param name="listInfoNode"></param>
		/// <param name="techViewName"></param>
		/// <returns></returns>
		private XmlElement CreateWhereClauseFromView(XmlNode listInfoNode, string techListName, string techViewName, WssSite.Views.Views views)
		{
			XmlElement query = listInfoNode.OwnerDocument.CreateElement("Query");
			if(techViewName == null)
			{
				// remove any filtering done by the view. For Issue type lists, this will also remove the IsCurrent filter
				query.InnerXml ="<WHERE><GT><FIELDREF Name='\"ID\"' /><VALUE Type='\"Counter\"'>0</VALUE></GT></WHERE>"; 
			}else
			{
				// Get View schema and remove Current = true part
				XmlNode ViewDescription =  views.GetView(techListName, techViewName);
				XmlElement DefaultQuery = (XmlElement)ViewDescription.SelectSingleNode("wss:Query", NameSpaceMgr);
				if(DefaultQuery == null)
				{
					query = null;
				}
				else
				{
					if(Grid != TimeGrid.None)
					{
						XmlElement CurrentClause = (XmlElement)DefaultQuery.SelectSingleNode("wss:Where/wss:Eq[wss:FieldRef/@Name = 'IsCurrent']", NameSpaceMgr);
						if(CurrentClause != null)
						{
							CurrentClause.ParentNode.RemoveChild(CurrentClause);	
						}
						// Remove the Where clause if it is empty
						XmlElement WhereClause = (XmlElement)DefaultQuery.SelectSingleNode("wss:Where", NameSpaceMgr);
						if(WhereClause != null && WhereClause.ChildNodes.Count == 0)
						{
							WhereClause.ParentNode.RemoveChild(WhereClause);	
						}
					}
				


					query.InnerXml = DefaultQuery.InnerXml;
				}
			}
			return query;
		}

		/// <summary>
		/// This converts a table with versions into a table with current versions for a number of time-points.
		/// For many timepoints, this may create a datatable that is significantly larger than the original data
		/// </summary>
		/// <param name="simpleTable"></param>
		/// <returns></returns>
		private DataTable ConvertToTimeGrid(DataTable simpleTable)
		{
			DataTable GridTable = new DataTable();
			foreach(DataColumn col in simpleTable.Columns)
			{
				GridTable.Columns.Add(col.ColumnName, col.DataType);
			}

			AddGridDate(GridTable);
			
			if(simpleTable.Rows.Count < 1)
			{
				return GridTable;
			}
			
			DataRow[] rows = simpleTable.Select("", "[Issue ID], [owshiddenversion] ");
			int rowCount = rows.Length;
			int index = rowCount -1;

			string CurrID = IDFromRow(rows[index]);

			DateTime FirstDate = (DateTime)simpleTable.Compute("min(Created)", "");
			DateTime LastDate = (DateTime)simpleTable.Compute("max(Modified)", "");

			MakeDatesFull(ref FirstDate, ref LastDate);

			// Empty ID means we are through the whole list
			while(CurrID != "")
			{
				// We loop through the Grid points backwards in time for every issue
				for( DateTime GridDate = LastDate; GridDate >= FirstDate; GridDate = SubtractTimestep(GridDate))
				{
					// we skip the rows until we find the first one that is created before the cuurent grid point
					while(index>=0 && CurrID == IDFromRow(rows[index]) && TimeFromRow(rows[index]) > GridDate)
					{
						index--;
					}
					// Out of data: exit
					if(index < 0)break;

					// If this is still about the same issue: we create a new row, otherwise, 
					// we should starts counting the grid points from the start
					if(CurrID == IDFromRow(rows[index]))
					{
						DataRow row = rows[index];
						ArrayList AllValues = new ArrayList(row.ItemArray);
						AllValues.Add(GridDate);
						GridTable.Rows.Add(AllValues.ToArray());
					}
					else
					{
						// new issue
						break;
					}

				}
				// loop through entries for previous issue
				while(index >= 0 && IDFromRow(rows[index]) == CurrID)
				{
					index--;
				}
				CurrID = index < 0 ? "" : IDFromRow(rows[index]);
			}

			GridTable.DefaultView.Sort = "GridDate";
			return GridTable;
		}

		private DateTime SubtractTimestep(DateTime gridDate)
		{
			switch(this.RuntimeGrid)
			{
				case TimeGrid.Daily:
					return gridDate.AddDays(-1);
				case TimeGrid.Weekly:
					return gridDate.AddDays(-7);
				case TimeGrid.Monthly:
					return gridDate.AddMonths(-1);
			}
			return gridDate.AddDays(-1);
		}

		private static DateTime TimeFromRow(DataRow row)
		{
			if(row == null) return DateTime.MinValue;
			DateTime result = DateTime.MinValue;
			if((string)row["Current"] == "-1")
			{
				result = (DateTime)row["Modified"];
			}else
			{
				result = (DateTime)row["Created"];
			}
			return result;
		}
		private static string IDFromRow(DataRow row)
		{
			if(row == null) return "";
			return (string)row["Issue ID"];
		}
		private void MakeDatesFull(ref DateTime FirstDate, ref DateTime LastDate)
		{
			FirstDate = FirstDate.Date;
			LastDate = LastDate.Date.AddDays(1);
			EvaluateAutoGrid(LastDate, FirstDate);
			switch(this.RuntimeGrid)
			{
				case TimeGrid.Daily:
					break;
				case TimeGrid.Weekly:
					LastDate = LastDate.AddDays(7-(int)LastDate.DayOfWeek);
					break;
				case TimeGrid.Monthly:
					LastDate = LastDate.AddDays(- (LastDate.Day -1));
					LastDate = LastDate.AddMonths(1);
					break;
				default:
					break;
			}
	
		}

		private void EvaluateAutoGrid(DateTime LastDate, DateTime FirstDate)
		{
			if(this.Grid == TimeGrid.Auto)
			{
				double Days = (LastDate - FirstDate	).TotalDays;
				if(Days < 40)
				{
					_calcGrid = TimeGrid.Daily;
				}else if(Days < 180)
				{
					_calcGrid = TimeGrid.Weekly;
				}else
				{
					_calcGrid = TimeGrid.Monthly;
				}
			}
		}

		private static void AddGridDate(DataTable GridTable)
		{
			GridTable.Columns.Add("GridDate", typeof(DateTime));
		}

		private Type TypeFromField(XmlElement field)
		{
			switch(field.GetAttribute("Type"))
			{
				case "DateTime": 
					return typeof(DateTime);
				case "Integer":
					return typeof(int);
				case "Number": 
					return typeof(float);
				default: 
					return typeof(string);
			}
		}

		private static Regex HtmlTagRegex = new Regex("<(?<close1>/?)(?<tag>\\w+)(\\s*\\w+\\s*=\\s*(\"[^\"]*\"|'[^']'|[^>]*))*\\s*/?>", RegexOptions.Singleline);
		private string ConvertToText(string html)
		{
			StringBuilder Text = new StringBuilder();
			MatchCollection Tags = HtmlTagRegex.Matches(html);
			int LastEnd = 0;
			foreach(Match Tag in Tags)
			{
				Text.Append(html.Substring(LastEnd, Tag.Index - LastEnd));
				bool Close = (Tag.Groups["close1"].Value != "");
				string TagName = Tag.Groups["tag"].Value.ToLower();
				if((TagName == "div" || TagName == "p") && Close)
				{
					Text.Append("\r\n");	
				}
				if(TagName == "br" || TagName == "ol" || TagName == "ul")
				{
					Text.Append("\r\n");	
				}
				if(TagName == "li" && !Close)
				{
					Text.Append("\r\n - ");
				}


				LastEnd = Tag.Index + Tag.Length;
			}
			Text = Text.Replace("&nbsp;", " ");
			Text = Text.Replace("&amp;", "&");
			Text = Text.Replace("&quot;", "\"");
			Text = Text.Replace("&apos;", "'");
			return Text.ToString();
		}

		private string FindTechListName(string listName, Lists lists, string site)
		{
			XmlNode ListCollectionNode = lists.GetListCollection();
			XmlElement List = (XmlElement)ListCollectionNode.SelectSingleNode(String.Format("wss:List[@Title='{0}']", listName), NameSpaceMgr);
			if(List == null)
			{
				throw new ArgumentException(String.Format("The list {0} could not be found in the site {1}", listName, site));
			}
			return List.GetAttribute("Name");
		}

		private string FindTechViewName(string viewName, string listName, Views views, string site)
		{
			if(viewName == null || viewName.Length == 0)return null;
			XmlNode ViewCollectionNode = views.GetViewCollection(listName);
			XmlElement View = (XmlElement)ViewCollectionNode.SelectSingleNode(String.Format("wss:View[@DisplayName='{0}']", viewName), NameSpaceMgr);
			if(View == null)
			{
				throw new ArgumentException(String.Format("The view {0} could not be found in the site {1}", viewName, site));
			}
			return View.GetAttribute("Name");
		}

		private static XmlNamespaceManager _nsmgr;
		private static XmlNamespaceManager NameSpaceMgr
		{
			get		
			{
				if(_nsmgr == null)
				{
                    _nsmgr = new XmlNamespaceManager(new NameTable());
					_nsmgr.AddNamespace("wss", "http://schemas.microsoft.com/sharepoint/soap/");
					_nsmgr.AddNamespace("s", "uuid:BDC6E3F0-6DA3-11d1-A2A3-00AA00C14882");
					_nsmgr.AddNamespace("dt", "uuid:C2F41010-65B3-11d1-A29F-00AA00C14882");
					_nsmgr.AddNamespace("rs", "urn:schemas-microsoft-com:rowset");
					_nsmgr.AddNamespace("z", "#RowsetSchema");


				}
				return _nsmgr;
			}
		}
	
		private ICredentials  GetCredential(Connection conn)
		{
			if(conn.IntegratedSecurity)
			{
				return CredentialCache.DefaultCredentials;
			}
			else
			{
				return new NetworkCredential(conn.UserName, conn.Password);
			}
		}
	}
}

