using System;
using System.Collections;
using System.Data;
using IDataReader = Microsoft.ReportingServices.DataProcessing.IDataReader;

namespace Macaw.ReportingServices.SharepointListExtension
{
	public class DataReader : IDataReader
	{
      
		private Connection _connection = null;
		private DataTable _table;
		private DataParameterCollection _parameters = null;

		/// <summary>
		/// DataReader should only be instantiated through the method ExecuteCommand on Command
		/// </summary>
		internal DataReader()
		{
		}

		/// <summary>
		/// DataReader should only be instantiated through the method ExecuteCommand on Command
		/// </summary>
		internal DataReader(Connection connection, DataTable table)
		{
			_connection  = connection;
			_table = table;
		}

		/// <summary>
		/// The DataReader is initialized with the DataTable that contains its data and schema information
		/// </summary>
		/// <param name="connection"></param>
		/// <param name="table"></param>
		/// <param name="parameters"></param>
		internal DataReader(Connection connection, DataTable table, DataParameterCollection parameters):this(connection, table)
		{
			_parameters = parameters;
		}


		IEnumerator TableEnum = null;

		/// <summary>
		/// Called repeatedly when the report is processed for each row
		/// </summary>
		/// <returns></returns>
		public bool Read()
		{
			if(TableEnum == null)
			{
				TableEnum = _table.Rows.GetEnumerator();
			}
			return TableEnum.MoveNext();
		}

		protected DataRow CurrentRow
		{
			get
			{
				return (DataRow)TableEnum.Current;
			}
		}

		/// <summary>
		/// Returns the number of columns in the table
		/// </summary>
		public int FieldCount
		{
			// Return the count of the number of columns, which in
			// this case is the size of the column metadata
			// array.
			get { return _table.Columns.Count; }
		}

		/// <summary>
		/// Returns the name of the column
		/// </summary>
		/// <param name="i"></param>
		/// <returns></returns>
		public string GetName(int i)
		{
			return _table.Columns[i].ColumnName;
		}
		/// <summary>
		/// Returns the column type
		/// </summary>
		/// <param name="i"></param>
		/// <returns></returns>
		public Type GetFieldType(int i)
		{
			// Return the actual Type class for the data type.
			return _table.Columns[i].DataType;
		}
		/// <summary>
		/// Returns the column value
		/// </summary>
		/// <param name="i"></param>
		/// <returns></returns>
		public Object GetValue(int i)
		{
			return CurrentRow[i];
		}
		/// <summary>
		/// Returns the column ordinal position
		/// </summary>
		/// <param name="name"></param>
		/// <returns></returns>
		public int GetOrdinal(string name)
		{
			foreach(DataColumn col in _table.Columns)
			{
				if(col.ColumnName == name)
				{
					return col.Ordinal;
				}
			}
			return -1;
		}

		public void Dispose()
		{
			_table.Dispose();
		}

	}
}

